#!/usr/bin/env python

import os
import sys
import optparse
import numpy as np
import pandas as pd
import json  #for writing dumps and combining stuffs
#import StringIO

#global variables
personLogFile = None
personIDLogFile = None
personTripFile = None
MAX_STEPS = 36000 #assuming that this would be the maximum simulation steps
STAGE_CYCLES = 5 #assumes that a stage is at least 5 steps long, so, capture only after a stage cycle

# class MyError is derived from super class Exception 
#class FatalTraCIError(Exception): 
#  
#    # Constructor or Initializer 
#    def __init__(self, value): 
#        self.value = value 
#  
#    # __str__ is to print() the value 
#    def __str__(self): 
#        return(repr(self.value)) 

# we need to import some python modules from the $SUMO_HOME/tools directory
if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)    
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")


from sumolib import checkBinary  # Checks for the binary in environ vars
import traci


def get_options():
    opt_parser = optparse.OptionParser()
    opt_parser.add_option("--nogui", action="store_true",
                         default=False, help="run the commandline version of sumo")
    options, args = opt_parser.parse_args()
    return options

def openLogFiles():
    global personLogFile, personIDLogFile, personTripFile
    personLogFile = open("person-logs.txt", "w")
    personIDLogFile = open("person-id-logs.txt", "w")
    personTripFile = open("person-trip-logs.csv", "w")
    header = "Person_Id" + "," + "Source_Stop_Id" + "," +  "Dest_Stop_Id" + "," + "Route_Id" + "," + "Vehicle_Id"
    personTripFile.write(header+"\n")
    
# contains TraCI control loop
def run():
    step = 0
#    personFound = 0    
    global personLogFile, MAX_STEPS, STAGE_CYCLES, personIDLogFile, personTripFile
#    personAList = list(np.linspace(1000,1010,10,dtype=int))       
    
    rideTrips = set()
    
    #main simulation loop begins
    while traci.simulation.getMinExpectedNumber() > 0: #simulate until all vehicles arrive
        
        traci.simulationStep() #capture the steps of simulation, one step at a time 
        step = step + 1
#        print("simulation time in seconds",traci.simulation.getCurrentTime()/60,"simulation step:",step)        
        
        if step % STAGE_CYCLES != 0: #only record once in a STAGE_CYCLES
            continue
        
        strLog = ""
                       
        peopleIDs = traci.person.getIDList() #Returns a list of person IDs       
        
        for i in range(len(peopleIDs)):
                #print("PersonID:",peopleIDs[i],"stage:",traci.person.getStage(peopleIDs[i]))
            #if len(traci.person.getVehicle(peopleIDs[i])) != 0: # if not empty
            currStage = traci.person.getStage(peopleIDs[i])
            
            if currStage.type == 3: #record only if driving or riding 
                #print("PersonID:"d,peopleIDs[i],"stage:",traci.person.getStage(peopleIDs[i]))                    
                #print("PersonID:",peopleIDs[i],"vehicle:",traci.person.getVehicle(peopleIDs[i]))
                personIDLogFile.write("PersonID: "+str(peopleIDs[i])+" StageType: "+ str(currStage.type) +"\n")
                vehicleID = traci.person.getVehicle(peopleIDs[i])
                
#                prevStage = traci.person.getStage(peopleIDs[i],-1) 
                try:
                    prevStage = traci.person.getStage(peopleIDs[i],-1) 
#                except FatalTraCIError:
#                    print("Previous Stage Error for Person",peopleIDs[i])
#                    continue
                except:
                    print("Error in Previous Stage retrieval, so, skip this personID: ", peopleIDs[i])
                    continue
                                
                if prevStage == None or currStage == None or vehicleID == None or vehicleID == "":
                    continue #do not record it
                strLog = strLog + "PersonID: %s" % (peopleIDs[i])
                strLog = strLog + "\t SourceStop: %s" % (prevStage.destStop)
                strLog = strLog + "\t DestStop: %s" % (currStage.destStop)
                strLog = strLog + "\t LineNo: %s" % (currStage.line)
#                strLog = strLog + "\t VehicleNo: %s" % (vehicleID) + "\t Vehilce TripID: %s" (traci.vehicle.getParameter(vehicleID,"tripId")) + "\n" 
                strLog = strLog + "\t VehicleNo: %s" % (vehicleID) + "\n" 
                
                #strTrip = str(peopleIDs[i]) + "," + str(prevStage.destStop).split(sep='#')[1] + "," + str(currStage.destStop).split(sep='#')[1] + "," + str(currStage.line) + "," + str(vehicleID).split(sep='_')[1]
                #personTripFile.write(strTrip + "\n")
                rideTrips.add(str(peopleIDs[i]) + "," + str(prevStage.destStop).split(sep='#')[1] + "," + str(currStage.destStop).split(sep='#')[1] + "," + str(currStage.line) + "," + str(vehicleID).split(sep='_')[1])
                              
#                strLog = strLog + "PersonID: %s, current stage: %s, vehicle: %s, previous stage: %s" % (peopleIDs[i],traci.person.getStage(peopleIDs[i]),traci.person.getVehicle(peopleIDs[i]),) + "\n"
                if i % 100 == 0: #flush the output if more people are there
                    personLogFile.write(strLog)
                    strLog = ""                    
                    
        
        if (step % 500) == 0 and len(peopleIDs) > 0:
            print("\nAt step:",step,"total public count:",len(peopleIDs))


        if strLog != "":
            personLogFile.write(strLog)
        
#        step += 1        #update simulation step
        if step > MAX_STEPS:
            break
    
    # while loop has ended        
    print("total steps:",step)    
    traci.close()   #end simulation
    
    print("Total trips on buses",len(rideTrips))
    for r in rideTrips:
        personTripFile.write(r + "\n")
    
    sys.stdout.flush()


def closeLogFiles():
    global personLogFile, personIDLogFile, personTripFile
    personLogFile.close()
    personIDLogFile.close()    
    personTripFile.close()


def getPreambles():
    ijPath = traci.simulation.findRoute("189[1][0]+20000","189[0]", "bus")
    print("Path from edge","189[1][0]+20000","to","78[1][1]","as per calculation is",ijPath,"\n")
#    
#    ijPath = traci.simulation.findRoute("200","85","bus")
#    print("Path from","58[1]","to","85","as per route calculation is",ijPath,"\n")
#    
#    ijPath = traci.simulation.findRoute("197","85","bus")
#    print("Path from","58[1]","to","85","as per route calculation is",ijPath,"\n")
    
    busStopList = traci.simulation.getBusStopIDList()
    print("bus stop IDs",busStopList)
    #create a dictionary of edge to bus stop map
    edge_busstop_map = dict()
    for b in busStopList:
        bEdge = traci.lane.getEdgeID(traci.busstop.getLaneID(b))
        edge_busstop_map[bEdge] = b
    print("Edge to bus stop mapping:\n",edge_busstop_map,"\n")
    with open('Edge2BusStop-Map.txt', 'w') as file:
        file.write(json.dumps(edge_busstop_map)) # use `json.loads` to do the reverse
    
    #compute the path for each bus stop pairs
    busStop_mat_column_names=['Source','Dest','Edges','TravelTime','TravelCost','Length','StopList']
    busStop_mat_df = pd.DataFrame(columns=busStop_mat_column_names)
    for i in busStopList:
        for j in busStopList:
            fromEdge = traci.lane.getEdgeID(traci.busstop.getLaneID(i))
            toEdge = traci.lane.getEdgeID(traci.busstop.getLaneID(j))
#            findRoute(self, fromEdge, toEdge, vType=''", depart=-1.0, routingMode=0)
            ijPath = traci.simulation.findRoute(fromEdge, toEdge, "bus")
            print("Path from",i,"to",j,ijPath)
            #compute reverse stop list
            rStopList = []
            for e in ijPath.edges:
                if e in edge_busstop_map.keys():
                    rStopList.append(edge_busstop_map[e])
            busStop_mat_df.loc[busStop_mat_df.shape[0]]=[i,j, ijPath.edges, ijPath.travelTime, ijPath.cost, ijPath.length,rStopList]                
    # saving the dataframe     
    busStop_mat_df.to_csv('busstop-mat-data.csv',index=False) 
    
    
    i = "busStop#31"
    j = "busStop#21"
    fromEdge = traci.lane.getEdgeID("189[1][0]+20000_1")
    toEdge = traci.lane.getEdgeID("134b_1")
#    findRoute(self, fromEdge, toEdge, vType=''", depart=-1.0, routingMode=0)
    ijPath = traci.simulation.findRoute(fromEdge, toEdge,"bus")
    print("Path from",i,"to",j,"as per program is",ijPath,"\n")
    
    print("\n",traci.simulation.findIntermodalRoute(fromEdge, toEdge,"bus","busStop#21"),"\n")
    print("\n",traci.simulation.findIntermodStagealRoute(fromEdge, toEdge),"\n\n")
    
    
    fromEdge = traci.lane.getEdgeID("134b_1")
    toEdge = traci.lane.getEdgeID("189[1][0]+20000_1")
#    findRoute(self, fromEdge, toEdge, vType=''", depart=-1.0, routingMode=0)
    ijPath = traci.simulation.findRoute(fromEdge, toEdge,"bus")    
    print("Path from",j,"to",i,"as per program is",ijPath,"\n")    
    print("Path edges",ijPath.edges,"\n")
    print("Path travel time",ijPath.travelTime,"\n")
    print("Path travel cost",ijPath.cost,"\n")
    print("Path travel length",ijPath.length,"\n")
    
    routeID = "14"
    edgeListOfRoute = traci.route.getEdges(routeID)
    print("Edge list of route",routeID,"is",edgeListOfRoute,"\n")
    
    print(edgeListOfRoute[0])
    print(edgeListOfRoute[len(edgeListOfRoute)-1])
    
    #generate trajectory meta-data for each route
    print("All route list:\n",traci.route.getIDList())   
    routeinfo_column_names=['RouteID','TravelTime','TravelCost','Length']
    routeinfo_df = pd.DataFrame(columns=routeinfo_column_names)
    for r in traci.route.getIDList():
        if len(traci.route.getEdges(r)) < 2:
            continue
        fromEdge = traci.route.getEdges(r)[0]
        toEdge = traci.route.getEdges(r)[len(traci.route.getEdges(r))-1]
        rPath = traci.simulation.findRoute(fromEdge, toEdge,"bus")
        routeinfo_df.loc[routeinfo_df.shape[0]]=[r, rPath.travelTime, rPath.cost, rPath.length]                
    # saving the dataframe     
    routeinfo_df.to_csv('route-info-data.csv',index=False) 

    #generate reverse trajectory meta-data for each route
    print("All route list:\n",traci.route.getIDList())   
    route_down_column_names=['RouteID','Edges','TravelTime','TravelCost','Length','StopList']
    route_down_df = pd.DataFrame(columns=route_down_column_names)
    for r in traci.route.getIDList():
        if len(traci.route.getEdges(r)) < 2:
            continue
        fromEdge = traci.route.getEdges(r)[len(traci.route.getEdges(r))-1]
        toEdge = traci.route.getEdges(r)[0]
        rPath = traci.simulation.findRoute(fromEdge, toEdge, "bus")
        if rPath == None or rPath.type != 3:
            continue
        if rPath.edges == (): #empty list
            #print("Reverse path has no edges")
            continue
        #compute reverse stop list
        rStopList = []
        for e in rPath.edges:
            if e in edge_busstop_map.keys:
                rStopList.append(edge_busstop_map[e])
        route_down_df.loc[route_down_df.shape[0]]=[r, rPath.edges, rPath.travelTime, rPath.cost, rPath.length,rStopList]                
    # saving the dataframe     
    route_down_df.to_csv('route-down-data.csv',index=False) 

def main():
    options = get_options()
    
    openLogFiles()
    
    # check binary
    if options.nogui:
        sumoBinary = checkBinary('sumo')
    else:
        sumoBinary = checkBinary('sumo-gui')
    sumoBinary = checkBinary('sumo')
    
    try:
        traci.load(['-c', 'run.sumo.cfg']) #only loads the simulation or scenario, not starts
        print("TraCI program loads the network successfully")
    except:
        print("TraCI program is unable to load the network")

    #traci starts sumo as a subprocess and then this script connects and runs
    #traci.start([sumoBinary, "-c", "demo.sumocfg","--tripinfo-output", "tripinfo.xml"])
    #traci.start([sumoBinary, "-c", "run.sumo.cfg", '--save-configuration', 'debug.sumocfg'])
    try:
        traci.start([sumoBinary, "-c", "run.sumo.cfg"]) #loads the scenario and takes only the zeroth step in simulation
        print("TraCI program starts the simulation successfully")
        
        #collect pre-simulation information
#        getPreambles()    
        
        #customized function to run each step of simulation and capture necessary data    
        run() 
    except:
        print("TraCI program is unable to start the simulation")             
    
#    if traci.hasGUI() == True:
#            print("This",traci.getVersion(),"version and GUI option is available")
    closeLogFiles()


# main entry point
if __name__ == "__main__":
    main()
    